# File: test_gpr.py
# File Created: Sunday, 21st July 2019 7:53:26 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
Tests for the GPR class
"""

import pytest


def TestGPR(object):
    # TODO
    pass